/*
 *	sleepwatcher
 *
 *	Copyright (c) 2002-2006 Bernhard Baehr
 *
 *	sleepwatcher.c - sleep mode watchdog program
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <ctype.h>
#include <errno.h>
#include <getopt.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <sys/wait.h>

#include <mach/mach_port.h>
#include <mach/mach_interface.h>
#include <mach/mach_init.h>

#include <CoreFoundation/CoreFoundation.h>
#include <CoreServices/CoreServices.h>

#include <IOKit/IOKitLib.h>
#include <IOKit/IOMessage.h>
#include <IOKit/hidsystem/IOHIDParameter.h>
#include <IOKit/hidsystem/IOHIDShared.h>
#include <IOKit/pwr_mgt/IOPMLib.h>


#define TIMER_RESOLUTION	0.1		/* seconds - when changing, adjust the man page! */

#define DENY_SLEEP		((char *) 1)	/* for args.allowsleepcommand */


static struct args {
	int		argc;
	char * const    *argv;
	char		*progname;
	int		verbose;
	int		daemon;
	char		*pidfile;
	char		*allowsleepcommand;
	char		*sleepcommand;
	char		*wakeupcommand;
	char		*displaydimcommand;
	char		*displayundimcommand;
	char		*displaysleepcommand;
	char		*displaywakeupcommand;
	long int	idletimeout;
	char		*idlecommand;
}	args;


static const struct option longopts[] = {
	{ "now",		no_argument,		NULL, 'n' },
	{ "version",		no_argument,		NULL, 'v' },
	{ "verbose",		no_argument,		NULL, 'V' },
	{ "daemon",		no_argument,		NULL, 'd' },
	{ "getidletime",	no_argument,		NULL, 'g' },
	{ "config",		required_argument,	NULL, 'f' },
	{ "pidfile",		required_argument,	NULL, 'p' },
	{ "allowsleep",		optional_argument,	NULL, 'a' },
	{ "sleep",		required_argument,	NULL, 's' },
	{ "wakeup",		required_argument,	NULL, 'w' },
	{ "displaydim",		required_argument,	NULL, 'D' },
	{ "displayundim",	required_argument,	NULL, 'E' },
	{ "displaysleep",	required_argument,	NULL, 'S' },
	{ "displaywakeup",	required_argument,	NULL, 'W' },
	{ "timeout",		required_argument,	NULL, 't' },
	{ "idle",		required_argument,	NULL, 'i' },
	{ NULL, 0, NULL, 0 }
};

#define GETOPT_STRING   "nvVdgf:p:a::s:w:D:E:S:W:t:i:"


/* Mac OS X 10.2 has basename() in libgen.h / libSystem.B.dylib, but Mac OS X 10.1 does not */
static char *basename (char *filename)
{
	char *p = strrchr(filename, '/');
	return (p ? p + 1 : filename);
}


static char *setstr (char *oldstr, char *newstr)
{
	if (oldstr)
		free (oldstr);
	return (newstr ? strdup(newstr) : NULL);
}


void message (int priority, const char *msg, ...)
{
	va_list ap;
	FILE    *out;
	
	if (args.verbose || priority < LOG_INFO) {
		va_start (ap, msg);
		if (args.daemon) {
			openlog (args.progname, LOG_PID, LOG_DAEMON);
			vsyslog (priority, msg, ap);
			closelog ();
		} else {
			out = (priority == LOG_INFO) ? stdout : stderr;
			fprintf (out, "%s: ", args.progname);
			vfprintf (out, msg, ap);
			fflush (out);
		}
	}
}


void writePidFile (char *pidfile)
{
	FILE    *fp;
	
	if (args.pidfile)
		unlink (args.pidfile);
	args.pidfile = setstr(args.pidfile, pidfile);
	if (args.pidfile) {
		fp = fopen(pidfile, "w");
		if (fp) {
			fprintf (fp, "%d", getpid());
			fclose (fp);
		} else
			message (LOG_ERR, "can't write pidfile %s\n", pidfile);
		
	}
}


static void usage (void)
{
        printf ("Usage: %s [-n] [-v] [-V] [-d] [-g] [-f configfile] [-p pidfile]\n"
		"		[-a[allowsleepcommand]]\n"
		"		[-s sleepcommand] [-w wakeupcommand]\n"
		"		[-D displaydimcommand] [-E displayundimcommand]\n"
		"		[-S displaysleepcommand] [-W displaywakeupcommand]\n"
		"		[-t timeout] [-i idlecommand]\n"
		"Daemon to monitor sleep, wakeup and idleness of the Mac\n"
		"-n or --now\n"
		"       sleep now and exit, ignoring other options\n"
		"-v or --version\n"
		"       display version and copyright information and exit\n"
		"-V or --verbose\n"
		"       log any action sleepwatcher performs\n"
		"-d or --daemon\n"
		"       run as a background daemon\n"
		"-g or -getidletime\n"
		"       write the time of no keyboard or mouse activity (in %g seconds)\n"
		"       and exit, ignoring other options\n"
		"-f or --config\n"
		"       read additional configuration parameters from configfile\n"
		"-p or --pidfile\n"
		"       write pidfile with the process id\n"
		"-a or --allowsleep\n"
		"       allow the Mac to sleep only when allowsleepcommand returns a zero\n"
		"       exit code, -a without allowsleepcommand denys sleeping\n"
		"       (note: no space between -a and the optional allowsleepcommand)\n"
                "-s or --sleep\n"
		"       execute sleepcommand when the Mac is put to sleep mode\n"
		"       sleepcommand must not take longer than 15 seconds because\n"
		"       after this timeout the sleep mode is forced by the system\n"
		"-w or --wakeup\n"
		"       execute wakeupcommand when the Mac wakes up\n"
		"-D or --displaydim\n"
		"       execute displaydimcommand when the display of the Mac is dimmed\n"
		"-E or --displayundim\n"
		"       execute displayundimcommand when the display of the Mac is undimmed\n"
		"       (without having gone to sleep)\n"
		"-S or --displaysleep\n"
		"       execute displaysleepcommand when the display of the Mac is put to\n"
		"       sleep mode\n"
		"-W or --displaywakeup\n"
		"       execute displaywakeupcommand when the display of the Mac wakes up\n"
		"-t or --timeout\n"
		"       set timeout for -i option (in %g seconds)\n"
		"-i or --idle\n"
		"       execute idlecommand when no user interaction (keyboard, mouse)\n"
		"       took place in the period given with the -t option\n",
		args.progname, TIMER_RESOLUTION, TIMER_RESOLUTION);
	exit (2);
}


static void copyright (void)
{
	printf ("sleepwatcher 2.0.5\n"
		"Copyright (C) 2002-2007 Bernhard Baehr (bernhard.baehr@gmx.de)\n"
		"This is free software that comes with ABSOLUTELY NO WARRANTY.\n"
		"See the GNU General Public License for details.\n");
	exit (2);
}


static int sleepImmediately (void)
{
	mach_port_t	masterPort;
	io_connect_t	rootPort;
	IOReturn	err;

	if (IOPMSleepEnabled()) {
		if ((err = IOMasterPort(MACH_PORT_NULL, &masterPort))) {
			fprintf (stderr, "%s: can't get mach master port: %ld\n", args.progname, (long) err);
			return (1);
		}
		rootPort = IOPMFindPowerManagement(masterPort);
		if (rootPort) {
			err = IOPMSleepSystem(rootPort);
			if (err) {
				fprintf (stderr, "%s: IOPMSleepSystem failed: %ld\n", args.progname, (long) err);
				return (1);
			}
		} else {
			fprintf (stderr, "%s: IOPMFindPowerManagement failed\n", args.progname);
			return (1);
		}
	} else {
		fprintf (stderr, "%s: sleep mode is disabled\n", args.progname);
		return (1);
	}
	return (0);
}


static long int getIdleTime (void)
/* returns mouse and keyboard idle time in TIMER_RESOLUTION seconds; returns -1 on error */
{
	mach_port_t		masterPort = 0;
	io_iterator_t		iter = 0;
	io_registry_entry_t     curObj = 0;
	CFMutableDictionaryRef  properties = NULL;
	CFTypeRef		obj = NULL;
	CFTypeID		type = 0;
	uint64_t		idletime = -1;

	if (IOMasterPort (MACH_PORT_NULL, &masterPort) != KERN_SUCCESS) {
		message (LOG_ERR, "can't get IOMasterPort\n");
		goto error;
	}
	IOServiceGetMatchingServices (masterPort, IOServiceMatching(kIOHIDSystemClass), &iter);
	if (iter == 0) {
		message (LOG_ERR, "can't access IOHIDSystem\n");
		goto error;
	}
	curObj = IOIteratorNext(iter);
	if (curObj == 0) {
		message (LOG_ERR, "got empty IOIterator\n");
		goto error;
	}
	if (IORegistryEntryCreateCFProperties(curObj, &properties, kCFAllocatorDefault, 0) != KERN_SUCCESS || ! properties) {
		message (LOG_ERR, "can't access HIDIdleTime\n");
		goto error;
	}
	obj = CFDictionaryGetValue(properties, CFSTR(kIOHIDIdleTimeKey));
	CFRetain (obj);
	type = CFGetTypeID(obj);
	if (type == CFDataGetTypeID())
		CFDataGetBytes ((CFDataRef) obj, CFRangeMake(0, sizeof(idletime)), (UInt8 *) &idletime);   
	else if (type == CFNumberGetTypeID())
		CFNumberGetValue ((CFNumberRef) obj, kCFNumberSInt64Type, &idletime);
	else { 
		message (LOG_ERR, "unsupported idle time data type\n", (int) type);
		goto error;
	}
	idletime /= 1000000000l * TIMER_RESOLUTION;      /* transform from 10**-9 to TIMER_RESOLUTION seconds */
error :
	if (masterPort)
		mach_port_deallocate (mach_task_self(), masterPort);
	if (obj)
		CFRelease(obj);
	if (curObj)
		IOObjectRelease (curObj);
	if (iter)
		IOObjectRelease (iter);
	if (properties)
		CFRelease ((CFTypeRef) properties);
	return (idletime);
}


static void readConfig (const char *configfile);	/* forward declaration */


static void setOption (const char c, char *optarg)
{

	char    *p;
	
	switch (c) {
		case 'n' :
			exit (sleepImmediately());
			break;
		case 'v' :
			copyright ();
			break;
		case 'V' :
			args.verbose = 1;
			break;
		case 'd' :
			args.daemon = 1;
			break;
		case 'g' :
			printf ("%ld\n", getIdleTime());
			exit (0);
			break;
		case 'f' :
			readConfig (optarg);
			break;
		case 'p' :
			writePidFile (optarg);
			break;
		case 'a' :
			if (args.allowsleepcommand == DENY_SLEEP)
				args.allowsleepcommand = NULL;
			args.allowsleepcommand = setstr(args.allowsleepcommand, optarg);
			if (args.allowsleepcommand == NULL)
				args.allowsleepcommand = DENY_SLEEP;
			break;
		case 's' :
			args.sleepcommand = setstr(args.sleepcommand, optarg);
			break;
		case 'w' :
			args.wakeupcommand = setstr(args.wakeupcommand, optarg);
			break;
		case 'D' :
			args.displaydimcommand = setstr(args.displaydimcommand, optarg);
			break;
		case 'E' :
			args.displayundimcommand = setstr(args.displayundimcommand, optarg);
			break;
		case 'S' :
			args.displaysleepcommand = setstr(args.displaysleepcommand, optarg);
			break;
		case 'W' :
			args.displaywakeupcommand = setstr(args.displaywakeupcommand, optarg);
			break;
		case 't' :
			args.idletimeout = 0;
			for (p = optarg; *p && isdigit(*p); p++)
				args.idletimeout = 10 * args.idletimeout + (*p - '0');
			if (*p) {
				message (LOG_ERR, "invalid digit(s) in timeout argument '%s'\n", optarg);
				args.idletimeout = -1;
			}
			break;
		case 'i' :
			args.idlecommand = setstr(args.idlecommand, optarg);
			break;
		default :
			usage ();
			break;
	}
}


static void readConfig (const char *configfile)
{
	char		buf[1024], *p, *q;
	struct option   const *op;
	int		l;
	
	FILE *fp = fopen(configfile, "r");
	if (! fp) {
		message (LOG_ERR, "can't read config file %s\n", configfile);
		return;
	}
	while (fgets(buf, sizeof(buf), fp)) {
		if (*buf == '#' || *buf == ';')		/* ignore comment lines */
			continue;
		if ((p = strchr(buf, '\n')))		/* remove newline */
			*p = '\0';
		for (p = q = buf; *p && *p != '='; isblank(*p++) || (*q++ = p[-1]))
			;				/* remove blanks before '=' */
		if (*p)
			*q++ = *p++;			/* '=' */
		while (*p && isblank(*p))		/* remove blanks before argument */
			p++;
		while (*q++ = *p++)			/* unmodified argument string */
			;
		for (op = longopts; op->name; op++) {
			l = strlen(op->name);
			if (! strncmp(op->name, buf, l)) {
				if (op->has_arg == no_argument && buf[l] ||
					op->has_arg != no_argument && buf[l] != '=')
					message (LOG_ERR, "malformed parameter '%s' in config file %s\n", buf, configfile);
				setOption (op->val, buf + l + 1);
				break;
			}
		}
		if (! op->name)
			message (LOG_ERR, "unknown parameter '%s' in config file %s\n", buf, configfile);
	}
	fclose (fp);
}


static void timerCallback (CFRunLoopTimerRef timer, void *info)
{
	long    idletime = getIdleTime();
	
/*
	fprintf (stderr, "timer %f\n", CFAbsoluteTimeGetCurrent());
*/
	if (idletime < args.idletimeout) {
		CFRunLoopTimerSetNextFireDate (timer, CFAbsoluteTimeGetCurrent() + (args.idletimeout - idletime) * TIMER_RESOLUTION);
	} else if (idletime <= args.idletimeout + 1) {
/*
		fprintf (stderr, "idle for %ld * %g seconds (timeout: %ld * %g seconds)\n",
			idletime, TIMER_RESOLUTION, args.idletimeout, TIMER_RESOLUTION);
*/
		message (LOG_INFO, "idle: %s: %d\n", args.idlecommand, system(args.idlecommand));
	}
/*
	else {
		fprintf (stderr, "idle for %ld * %g seconds, waiting for activity\n", idletime, TIMER_RESOLUTION);
	}
*/
}


static void installTimer (void)
{
	static CFRunLoopTimerRef	timer = NULL;
	
	if (args.idletimeout) {
		if (timer)
			CFRunLoopTimerSetNextFireDate (timer, CFAbsoluteTimeGetCurrent() + args.idletimeout * TIMER_RESOLUTION);
		else {
			timer = CFRunLoopTimerCreate(kCFAllocatorDefault,
				CFAbsoluteTimeGetCurrent() + args.idletimeout * TIMER_RESOLUTION,
				args.idletimeout * TIMER_RESOLUTION, 0, 0, timerCallback, NULL);
			CFRunLoopAddTimer (CFRunLoopGetCurrent(), timer, kCFRunLoopDefaultMode);
		}
	} else {
		if (timer) {
			CFRunLoopRemoveTimer (CFRunLoopGetCurrent(), timer, kCFRunLoopDefaultMode);
			timer = NULL;
		}
	}
}


static void parseArgs (int argc, char * const *argv)
{
	int	c, i;

	args.argc = argc;
	args.argv = argv;
	args.progname = basename(*argv);
	args.verbose = 0;
	args.daemon = 0;
	args.allowsleepcommand = setstr(args.allowsleepcommand, NULL);
	args.sleepcommand = setstr(args.sleepcommand, NULL);
	args.wakeupcommand = setstr(args.wakeupcommand, NULL);
	args.displaydimcommand = setstr(args.displaydimcommand, NULL);
	args.displayundimcommand = setstr(args.displayundimcommand, NULL);
	args.displaysleepcommand = setstr(args.displaysleepcommand, NULL);
	args.displaywakeupcommand = setstr(args.displaywakeupcommand, NULL);
	args.idletimeout = 0;
	args.idlecommand = setstr(args.idlecommand, NULL);
	writePidFile (NULL);
	if (argc == 1)
		usage ();
	optreset = optind = 1;
	while ((c = getopt_long(argc, argv, GETOPT_STRING, longopts, &i)) != -1)
		setOption (c, optarg);
	if (args.idletimeout == -1) {		/* invalid timeout value was specified */
		args.idletimeout = 0;
		args.idlecommand = setstr(args.idlecommand, NULL);
	}
	if (args.idletimeout == 0 && args.idlecommand) {
		message (LOG_ERR, "idlecommand without timeout ignored\n");
		args.idlecommand = setstr(args.idlecommand, NULL);
	}
	if (args.idletimeout && args.idlecommand == NULL) {
		message (LOG_ERR, "timeout without idlecommand ignored\n");
		args.idletimeout = 0;
	}
}


static void signalCallback (int sig)
{
	switch (sig) {
	case SIGHUP :
		message (LOG_INFO, "got SIGHUP - reconfiguring\n");	
		parseArgs (args.argc, args.argv);
		installTimer ();
		break;
	case SIGTERM :
	case SIGINT :
		message (LOG_INFO, "got %s - exiting\n", sig == SIGTERM ? "SIGTERM" : "SIGINT");	
		writePidFile (NULL);
		exit (0);
		break;
	}
}


void powerCallback (void *rootPort, io_service_t y, natural_t msgType, void *msgArgument)
{
	int	result;
	char	*s;

/*
	fprintf (stderr, "powerCallback: message_type %08lx, arg %08lx\n",
		(long unsigned int) msgType, (long  unsigned int) msgArgument);
*/
	switch (msgType) {
	case kIOMessageCanSystemSleep :
		if (args.allowsleepcommand == NULL) {
			result = 0;
			message (LOG_INFO, "allow sleep\n");
		} else if (args.allowsleepcommand == DENY_SLEEP) {
			result = -1;
			message (LOG_INFO, "deny sleep\n");
		} else {
			result = system(args.allowsleepcommand);
			s = (! WIFEXITED(result) || WEXITSTATUS(result)) ? "deny" : "allow";
			message (LOG_INFO, "%s sleep: %s: exit=%s status=%d\n", s, args.allowsleepcommand,
				WIFEXITED(result) ? "ok" : "err", WEXITSTATUS(result));
			result = ! WIFEXITED(result) || WEXITSTATUS(result);
		}
		if (result)
			IOCancelPowerChange (* (io_connect_t *) rootPort, (long) msgArgument);
		else
			IOAllowPowerChange (* (io_connect_t *) rootPort, (long) msgArgument);
		break;
	case kIOMessageSystemWillSleep :
		if (args.sleepcommand)
			message (LOG_INFO, "sleep: %s: %d\n", args.sleepcommand, system(args.sleepcommand));
		IOAllowPowerChange (* (io_connect_t *) rootPort, (long) msgArgument);
		break;
	case kIOMessageSystemHasPoweredOn :
		if (args.wakeupcommand)
			message (LOG_INFO, "wakeup: %s: %d\n", args.wakeupcommand, system(args.wakeupcommand));
		break;
	}
}


void displayCallback (void *refcon, io_service_t y, natural_t msgType, void *msgArgument)
{
	static enum { displayOn, displayDimmed, displayOff } state = displayOn;
	
/*
	fprintf (stderr, "displayCallback: message_type %08lx, arg %08lx\n",
		(long unsigned int) msgType, (long  unsigned int) msgArgument);
*/
	switch (msgType) {
	case kIOMessageDeviceWillPowerOff :
		state++;
		if (state == displayDimmed && args.displaydimcommand)
			message (LOG_INFO, "displaydim: %s: %d\n", args.displaydimcommand,
				system(args.displaydimcommand));
		else if (state == displayOff && args.displaysleepcommand)
			message (LOG_INFO, "displaysleep: %s: %d\n", args.displaysleepcommand,
				system(args.displaysleepcommand));
		break;
	case kIOMessageDeviceHasPoweredOn :
		if (state == displayDimmed && args.displayundimcommand)
			message (LOG_INFO, "displayundim: %s: %d\n", args.displayundimcommand,
				system(args.displayundimcommand));
		else if (args.displaywakeupcommand)
			message (LOG_INFO, "displaywakeup: %s: %d\n", args.displaywakeupcommand,
				system(args.displaywakeupcommand));
		state = displayOn;
		break;
	}
}


static void initializePowerNotifications (void)
{
	static io_connect_t	rootPort;	/* used by powerCallback() via refcon */
	
	IONotificationPortRef	notificationPort;
	io_object_t		notifier;

	rootPort = IORegisterForSystemPower(&rootPort, &notificationPort, powerCallback, &notifier);
	if (! rootPort) {
		message (LOG_ERR, "IORegisterForSystemPower failed\n");
		exit (1);
	}
	CFRunLoopAddSource (CFRunLoopGetCurrent(), IONotificationPortGetRunLoopSource(notificationPort), kCFRunLoopDefaultMode);
}


static void initializeDisplayNotifications (void)
{
	io_service_t		displayWrangler;
	IONotificationPortRef	notificationPort;
	io_object_t		notifier;

	displayWrangler = IOServiceGetMatchingService(kIOMasterPortDefault, IOServiceNameMatching("IODisplayWrangler"));
	if (! displayWrangler) {
		message (LOG_ERR, "IOServiceGetMatchingService failed\n");
		exit (1);
	}
	notificationPort = IONotificationPortCreate(kIOMasterPortDefault);
	if (! notificationPort) {
		message (LOG_ERR, "IONotificationPortCreate failed\n");
		exit (1);
	}
	if (IOServiceAddInterestNotification(notificationPort, displayWrangler, kIOGeneralInterest,
		displayCallback, NULL, &notifier) != kIOReturnSuccess) {
		message (LOG_ERR, "IOServiceAddInterestNotification failed\n");
		exit (1);
	}
	CFRunLoopAddSource (CFRunLoopGetCurrent(), IONotificationPortGetRunLoopSource(notificationPort), kCFRunLoopDefaultMode);
	IOObjectRelease (displayWrangler);
}


int main (int argc, char * const *argv)
{
	parseArgs (argc, argv);
	if (args.daemon) {
		if (daemon(0, 0)) {
			message (LOG_ERR, "daemonizing failed: %d\n", errno);
			return (1);
		}
		writePidFile (args.pidfile);
		fclose (stdin);
		fclose (stdout);
		fclose (stderr);
	}
	signal (SIGHUP, signalCallback);
	signal (SIGINT, signalCallback);
	signal (SIGTERM, signalCallback);
	installTimer ();
	initializePowerNotifications ();
	initializeDisplayNotifications ();
	CFRunLoopRun ();
	return (0);
}

